@@ -2,8 +2,8 @@
 #include <malloc.h>
 #include <errno.h>
 #include <msvcrt.h>
+#include <stdio.h>
 
-int __cdecl vsprintf (char *, const char *, va_list);
 int __cdecl vsprintf_s (char *, size_t, const char *, va_list);
 static int __cdecl _int_vsprintf_s (char *, size_t, const char *, va_list);
 static int __cdecl _stub (char *, size_t, const char *, va_list);
@@ -19,7 +19,7 @@ _stub (char *_DstBuf, size_t _Size, const char *_Format, va_list _ArgList)
   if (f == _stub)
     {
 	f = (int __cdecl (*)(char *, size_t, const char *, va_list))
-	    GetProcAddress (__mingw_get_msvcrt_handle (), "vprintf_s");
+	    GetProcAddress (__mingw_get_msvcrt_handle (), "vsprintf_s");
 	if (!f)
 	  f = _int_vsprintf_s;
 	__MINGW_IMP_SYMBOL(vsprintf_s) = f;
@@ -34,7 +34,7 @@ vsprintf_s (char *_DstBuf, size_t _Size, const char *_Format, va_list _ArgList)
 }
 
 static int __cdecl
-_int_vsprintf_s (char *_DstBuf, size_t __UNUSED_PARAM(_Size), const char *_Format, va_list _ArgList)
+_int_vsprintf_s (char *_DstBuf, size_t _Size, const char *_Format, va_list _ArgList)
 {
-  return vsprintf (_DstBuf, _Format, _ArgList);
+  return __ms_vsnprintf (_DstBuf, _Size, _Format, _ArgList);
 }
