@@ -77,13 +77,13 @@
 
 /* FIXME: The following belongs in values.h, but current MinGW
  * has nothing useful there!  OTOH, values.h is not a standard
- * header, and it's use may be considered obsolete; perhaps it
+ * header, and its use may be considered obsolete; perhaps it
  * is better to just keep these definitions here.
  */
 
 #include <pshpack1.h>
 /* workaround gcc bug */
-#ifdef __GNUC__
+#if defined(__GNUC__) && !defined(__clang__)
 #define ATTRIB_GCC_STRUCT __attribute__((gcc_struct))
 #else
 #define ATTRIB_GCC_STRUCT
@@ -182,56 +182,6 @@ typedef union ATTRIB_GCC_STRUCT __uI128 {
  */
 #define PFORMAT_INFNAN      -32768
 
-#ifdef _WIN32
-/*
- * The Microsoft standard for printing `%e' format exponents is
- * with a minimum of three digits, unless explicitly set otherwise,
- * by a prior invocation of the `_set_output_format()' function.
- *
- * The following macro allows us to replicate this behaviour.
- */
-# define PFORMAT_MINEXP    __pformat_exponent_digits()
- /*
-  * However, this feature is unsupported for versions of the
-  * MSVC runtime library prior to msvcr80.dll, and by default,
-  * MinGW uses an earlier version, (equivalent to msvcr60.dll),
-  * for which `_TWO_DIGIT_EXPONENT' will be undefined.
-  */
-# ifndef _TWO_DIGIT_EXPONENT
- /*
-  * This hack works around the lack of the `_set_output_format()'
-  * feature, when supporting versions of the MSVC runtime library
-  * prior to msvcr80.dll; it simply enforces Microsoft's original
-  * convention, for all cases where the feature is unsupported.
-  */
-#  define _get_output_format()  0
-#  define _TWO_DIGIT_EXPONENT   1
-# endif
-/*
- * Irrespective of the MSVCRT version supported, *we* will add
- * an additional capability, through the following inline function,
- * which will allow the user to choose his own preferred default
- * for `PRINTF_EXPONENT_DIGITS', through the simple expedient
- * of defining it as an environment variable.
- */
-static
-int __pformat_exponent_digits( void )
-{
-  char *exponent_digits = getenv( "PRINTF_EXPONENT_DIGITS" );
-  return ((exponent_digits != NULL) && ((unsigned)(*exponent_digits - '0') < 3))
-    || (_get_output_format() & _TWO_DIGIT_EXPONENT)
-    ? 2
-    : 3
-    ;
-}
-#else
-/*
- * When we don't care to mimic Microsoft's standard behaviour,
- * we adopt the C99/POSIX standard of two digit exponents.
- */
-# define PFORMAT_MINEXP         2
-#endif
-
 typedef union
 {
   /* A data type agnostic representation,
@@ -460,6 +410,7 @@ void __pformat_putc( int c, __pformat_t *stream )
 static
 void __pformat_putchars( const char *s, int count, __pformat_t *stream )
 {
+#ifndef __BUILD_WIDEAPI
   /* Handler for `%c' and (indirectly) `%s' conversion specifications.
    *
    * Transfer characters from the string buffer at `s', character by
@@ -503,7 +454,56 @@ void __pformat_putchars( const char *s, int count, __pformat_t *stream )
 
   /* Emit the data...
    */
-#ifdef __BUILD_WIDEAPI
+  while( count-- )
+    /*
+     * copying the requisite number of characters from the input.
+     */
+    __pformat_putc( *s++, stream );
+
+  /* If we still haven't consumed the entire specified field width,
+   * we must be doing flush left justification; any residual width
+   * must be filled with blanks, to the right of the output value.
+   */
+  while( stream->width-- > 0 )
+    __pformat_putc( '\x20', stream );
+
+#else  /* __BUILD_WIDEAPI */
+
+  int len;
+
+  if( (stream->precision >= 0) && (count > stream->precision) )
+    count = stream->precision;
+
+  if( (stream->flags & PFORMAT_TO_FILE) && (stream->flags & PFORMAT_NOLIMIT) )
+  {
+    int __cdecl __ms_fwprintf(FILE *, const wchar_t *, ...);
+
+    if( stream->width > count )
+    {
+      if( (stream->flags & PFORMAT_LJUSTIFY) == 0 )
+        len = __ms_fwprintf( (FILE *)(stream->dest), L"%*.*S", stream->width, count, s );
+      else
+        len = __ms_fwprintf( (FILE *)(stream->dest), L"%-*.*S", stream->width, count, s );
+    }
+    else
+    {
+      len = __ms_fwprintf( (FILE *)(stream->dest), L"%.*S", count, s );
+    }
+    if( len > 0 )
+      stream->count += len;
+    stream->width = PFORMAT_IGNORE;
+    return;
+  }
+
+  if( stream->width > count )
+    stream->width -= count;
+  else
+    stream->width = PFORMAT_IGNORE;
+
+  if( (stream->width > 0) && ((stream->flags & PFORMAT_LJUSTIFY) == 0) )
+    while( stream->width-- )
+      __pformat_putc( '\x20', stream );
+
   {
     /* mbrtowc */
     size_t l;
@@ -526,20 +526,11 @@ void __pformat_putchars( const char *s, int count, __pformat_t *stream )
       __pformat_putc((int)w[0], stream);
     }
   }
-#else
-  while( count-- )
-    /*
-     * copying the requisite number of characters from the input.
-     */
-    __pformat_putc( *s++, stream );
-#endif
 
-  /* If we still haven't consumed the entire specified field width,
-   * we must be doing flush left justification; any residual width
-   * must be filled with blanks, to the right of the output value.
-   */
   while( stream->width-- > 0 )
     __pformat_putc( '\x20', stream );
+
+#endif  /* __BUILD_WIDEAPI */
 }
 
 static
@@ -558,12 +549,17 @@ void __pformat_puts( const char *s, __pformat_t *stream )
    * (after first verifying that the input pointer is not NULL).
    */
   if( s == NULL ) s = "(null)";
-  __pformat_putchars( s, strlen( s ), stream );
+
+  if( stream->precision >= 0 )
+    __pformat_putchars( s, strnlen( s, stream->precision ), stream );
+  else
+    __pformat_putchars( s, strlen( s ), stream );
 }
 
 static
 void __pformat_wputchars( const wchar_t *s, int count, __pformat_t *stream )
 {
+#ifndef __BUILD_WIDEAPI
   /* Handler for `%C'(`%lc') and `%S'(`%ls') conversion specifications;
    * (this is a wide character variant of `__pformat_putchars()').
    *
@@ -608,27 +604,67 @@ void __pformat_wputchars( const wchar_t *s, int count, __pformat_t *stream )
   /* Emit the data, converting each character from the wide
    * to the multibyte domain as we go...
    */
-#ifdef __BUILD_WIDEAPI
-  len = count;
-  while(len-- > 0 && *s != 0)
-  {
-      __pformat_putc(*s++, stream);
-  }
-  count = len;
-#else
   while( (count-- > 0) && ((len = wcrtomb( buf, *s++, &state )) > 0) )
   {
     char *p = buf;
     while( len-- > 0 )
       __pformat_putc( *p++, stream );
   }
-#endif
+
   /* If we still haven't consumed the entire specified field width,
    * we must be doing flush left justification; any residual width
    * must be filled with blanks, to the right of the output value.
    */
   while( stream->width-- > 0 )
     __pformat_putc( '\x20', stream );
+
+#else  /* __BUILD_WIDEAPI */
+
+  int len;
+
+  if( (stream->precision >= 0) && (count > stream->precision) )
+    count = stream->precision;
+
+  if( (stream->flags & PFORMAT_TO_FILE) && (stream->flags & PFORMAT_NOLIMIT) )
+  {
+    int __cdecl __ms_fwprintf(FILE *, const wchar_t *, ...);
+
+    if( stream->width > count )
+    {
+      if( (stream->flags & PFORMAT_LJUSTIFY) == 0 )
+        len = __ms_fwprintf( (FILE *)(stream->dest), L"%*.*s", stream->width, count, s );
+      else
+        len = __ms_fwprintf( (FILE *)(stream->dest), L"%-*.*s", stream->width, count, s );
+    }
+    else
+    {
+      len = __ms_fwprintf( (FILE *)(stream->dest), L"%.*s", count, s );
+    }
+    if( len > 0 )
+      stream->count += len;
+    stream->width = PFORMAT_IGNORE;
+    return;
+  }
+
+  if( stream->width > count )
+    stream->width -= count;
+  else
+    stream->width = PFORMAT_IGNORE;
+
+  if( (stream->width > 0) && ((stream->flags & PFORMAT_LJUSTIFY) == 0) )
+    while( stream->width-- )
+      __pformat_putc( '\x20', stream );
+
+  len = count;
+  while(len-- > 0 && *s != 0)
+  {
+      __pformat_putc(*s++, stream);
+  }
+
+  while( stream->width-- > 0 )
+    __pformat_putc( '\x20', stream );
+
+#endif  /* __BUILD_WIDEAPI */
 }
 
 static
@@ -648,7 +684,11 @@ void __pformat_wcputs( const wchar_t *s, __pformat_t *stream )
    * (after first verifying that the input pointer is not NULL).
    */
   if( s == NULL ) s = L"(null)";
-  __pformat_wputchars( s, wcslen( s ), stream );
+
+  if( stream->precision >= 0 )
+    __pformat_wputchars( s, wcsnlen( s, stream->precision ), stream );
+  else
+    __pformat_wputchars( s, wcslen( s ), stream );
 }
 
 static
@@ -1460,7 +1500,11 @@ void __pformat_emit_efloat( int sign, char *value, int e, __pformat_t *stream )
     exp_width++;
 
   /* Ensure that this is at least as many as the standard requirement.
+   * The C99 standard requires the expenent to contain at least two
+   * digits, unless specified explicitly otherwise.
    */
+  if (stream->expmin == -1)
+    stream->expmin = 2;
   if( exp_width < stream->expmin )
     exp_width = stream->expmin;
 
@@ -1992,8 +2036,8 @@ void __pformat_emit_xfloat( __pformat_fpreg_t value, __pformat_t *stream )
   {
     /* taking the rightmost digit in each pass...
      */
-    int c = value.__pformat_fpreg_mantissa & 0xF;
-    if( c == (int) value.__pformat_fpreg_mantissa)
+    unsigned c = value.__pformat_fpreg_mantissa & 0xF;
+    if( c == value.__pformat_fpreg_mantissa)
     {
       /* inserting the radix point, when we reach the last,
        * (i.e. the most significant digit), unless we found no
@@ -2273,9 +2317,14 @@ __pformat (int flags, void *dest, int max, const APICHAR *fmt, va_list argv)
     (wchar_t)(0),				/* leave it unspecified	      */
     0,						/* zero output char count     */
     max,					/* establish output limit     */
-    PFORMAT_MINEXP				/* exponent chars preferred   */
+    -1						/* exponent chars preferred;
+                                                   -1 means to be determined. */
   };
 
+#ifdef __BUILD_WIDEAPI
+  const APICHAR *literal_string_start = NULL;
+#endif
+
   format_scan: while( (c = *fmt++) != 0 )
   {
     /* Format string parsing loop...
@@ -2300,6 +2349,15 @@ __pformat (int flags, void *dest, int max, const APICHAR *fmt, va_list argv)
        */
       int *width_spec = &stream.width;
 
+  #ifdef __BUILD_WIDEAPI
+      if (literal_string_start)
+      {
+        stream.width = stream.precision = PFORMAT_IGNORE;
+        __pformat_wputchars( literal_string_start, fmt - literal_string_start - 1, &stream );
+        literal_string_start = NULL;
+      }
+  #endif
+
       /* Reset initial state for flags, width and precision specs...
        */
       stream.flags = flags;
@@ -2324,7 +2382,12 @@ __pformat (int flags, void *dest, int max, const APICHAR *fmt, va_list argv)
 	     * on GNU/Linux does NOT appear to require this, but POSIX
 	     * and SUSv3 do seem to demand it).
 	     */
+    #ifndef __BUILD_WIDEAPI
 	    __pformat_putc( c, &stream );
+    #else
+        stream.width = stream.precision = PFORMAT_IGNORE;
+        __pformat_wputchars( L"%", 1, &stream );
+    #endif
 	    goto format_scan;
 
 	  case 'C':
@@ -2852,8 +2915,8 @@ __pformat (int flags, void *dest, int max, const APICHAR *fmt, va_list argv)
 	       */
 	      length = PFORMAT_LENGTH_LONG;
 
-	      state = PFORMAT_END;
-	      break;
+	    state = PFORMAT_END;
+	    break;
 
 	  case 'L':
 	    /*
@@ -3080,7 +3143,12 @@ __pformat (int flags, void *dest, int max, const APICHAR *fmt, va_list argv)
 	       * backtrack, and emit it as literal text...
 	       */
 	      fmt = backtrack;
+      #ifndef __BUILD_WIDEAPI
 	      __pformat_putc( '%', &stream );
+      #else
+          stream.width = stream.precision = PFORMAT_IGNORE;
+          __pformat_wputchars( L"%", 1, &stream );
+      #endif
 	      goto format_scan;
 	    }
 	}
@@ -3091,12 +3159,25 @@ __pformat (int flags, void *dest, int max, const APICHAR *fmt, va_list argv)
       /* We just parsed a character which is not included within any format
        * specification; we simply emit it as a literal.
        */
+  #ifndef __BUILD_WIDEAPI
       __pformat_putc( c, &stream );
+  #else
+      if (literal_string_start == NULL)
+        literal_string_start = fmt - 1;
+  #endif
   }
 
   /* When we have fully dispatched the format string, the return value is the
    * total number of bytes we transferred to the output destination.
    */
+#ifdef __BUILD_WIDEAPI
+  if (literal_string_start)
+  {
+    stream.width = stream.precision = PFORMAT_IGNORE;
+    __pformat_wputchars( literal_string_start, fmt - literal_string_start - 1, &stream );
+  }
+#endif
+
   return stream.count;
 }
 
